/*
* msgop.c: Illustrate the msgsnd() and msgrcv() functions.
*
* This is a simple exerciser of the message send and receive
* routines. It allows the user to attempt to send and receive as
many
* messages as wanted to or from one message queue.
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

static int ask();
char first_on_queue[] = "-> first message on queue",
     full_buf[] = "Message buffer overflow. Extra message text discarded.";


int main()
{
   int c; /* message text input */
   int choice; /* users selected operation code */
   int i; /* loop control for mtext */
   int msgflg; /* message flags for the operation */
   struct msgbuf *msgp; /* pointer to the message buffer */
   int msgsz; /* message size */
   long msgtyp; /* desired message type */
   int msqid, /* message queue ID to be used */
   maxmsgsz, /* size of allocated message buffer */
   rtrn; /* return value from msgrcv or msgsnd */

   fprintf(stderr, "Enter msqid: ");
   scanf("%i", &msqid);
  
   fprintf(stderr,"Enter the message buffer size you want:");
   scanf("%i", &maxmsgsz);
   if (maxmsgsz < 0) {
      fprintf(stderr, "msgop: %s\n", "The message buffer size must be >= 0.");
      _exit(1);
   }
   msgp = (struct msgbuf *)malloc((unsigned)(sizeof(struct msgbuf) - 
           sizeof msgp->mtext + maxmsgsz));

   if (msgp == NULL) {
     fprintf(stderr, "msgop: %s %d byte messages.\n",
             "could not allocate message buffer for", maxmsgsz);
     _exit(1);
   }

   /* Loop through message operations until the user is ready to quit. */
   while (choice = ask()) {
      switch (choice) {
	case 1: 
	  fprintf(stderr, "Enter msgp->mtype: ");
 	  scanf("%li", &msgp->mtype);
	  if (maxmsgsz) {
          fprintf(stderr, "Enter a one line message:\n");
    	    gets(full_buf);   
          strcpy(msgp->mtext,full_buf);
	    msgsz = strlen(full_buf);
  
 	  fprintf(stderr,"\nMeaningful msgsnd flag is:\n");
        fprintf(stderr, "\tIPC_NOWAIT =\t%#8.8o\n",IPC_NOWAIT);
        fprintf(stderr, "Enter msgflg: ");
        scanf("%i", &msgflg);
        fprintf(stderr, "%s(%d, msgp, %d, %#o)\n", "msgop: Calling msgsnd", msqid, msgsz, msgflg);
 	fprintf(stderr, "msgp->mtype = %ld\n", msgp->mtype);
 	fprintf(stderr, "msgp->mtext = \"");
	for (i = 0; i < msgsz; i++)
 	fputc(msgp->mtext[i], stderr);
 	fprintf(stderr, "\"\n");
	rtrn = msgsnd(msqid, msgp, msgsz, msgflg);
	if (rtrn == -1)
	perror("msgop: msgsnd failed");
	else
 	fprintf(stderr,"Enter the message buffer size you want:");
	
 	scanf("%i", &maxmsgsz);
	if (maxmsgsz < 0) {
 	fprintf(stderr, "msgop: %s\n",
	"The message buffer size must be >= 0.");
	exit(1);
}
 fprintf(stderr,"msgop: msgsnd returned %d\n", rtrn);
break;
case 2: /* msgrcv() requested: Get the arguments, make the
call, and report the results. */
for (msgsz = -1; msgsz < 0 || msgsz > maxmsgsz;
 scanf("%i", &msgsz))
 fprintf(stderr, "%s (0 <= msgsz <= %d): ",
"Enter msgsz", maxmsgsz);
 fprintf(stderr, "msgtyp meanings:\n");
 fprintf(stderr, "\t 0 %s\n", first_on_queue);
 fprintf(stderr, "\t>0 %s of given type\n",
first_on_queue);
 fprintf(stderr, "\t<0 %s with type <= |msgtyp|\n",
first_on_queue);
 fprintf(stderr, "Enter msgtyp: ");
 scanf("%li", &msgtyp);
 fprintf(stderr,
"Meaningful msgrcv flags are:\n");
 fprintf(stderr, "\tMSG_NOERROR =\t%#8.8o\n",
MSG_NOERROR);
 fprintf(stderr, "\tIPC_NOWAIT =\t%#8.8o\n",
IPC_NOWAIT);
 fprintf(stderr, "Enter msgflg: ");
 scanf("%i", &msgflg);
 fprintf(stderr, "%s(%d, msgp, %d, %ld, %#o);\n",
"msgop: Calling msgrcv", msqid, msgsz,
msgtyp, msgflg);
rtrn = msgrcv(msqid, msgp, msgsz, msgtyp, msgflg);
if (rtrn == -1)
perror("msgop: msgrcv failed");
else {
 fprintf(stderr, "msgop: %s %d\n",
"msgrcv returned", rtrn);
 fprintf(stderr, "msgp->mtype = %ld\n",
msgp->mtype);
 fprintf(stderr, "msgp->mtext is: \"");
for (i = 0; i < rtrn; i++)
 fputc(msgp->mtext[i], stderr);
 fprintf(stderr, "\"\n");
}
break;
default:
 fprintf(stderr, "msgop: operation unknown\n");
break;
}
}
exit(0);
}
/*
* Ask the user what to do next. Return the users choice code.
* Dont return until the user selects a valid choice.
*/
static
ask()
{
int response; /* Users response. */
do {
 fprintf(stderr, "Your options are:\n");
 fprintf(stderr, "\tExit =\t0 or Control-D\n");

 fprintf(stderr, "\tmsgsnd =\t1\n");
 fprintf(stderr, "\tmsgrcv =\t2\n");
 fprintf(stderr, "Enter your choice: ");
/* Preset response so "^D" will be interpreted as exit. */
response = 0;
 scanf("%i", &response);
} while (response < 0 || response > 2);
return(response);
}
