/******************************************
MODULO: semop.c
SCOPO: Illustrare il funz. di semop()
******************************************/
#include    <stdio.h>
#include    <stdlib.h>
#include    <sys/types.h>
#include    <sys/ipc.h>
#include    <sys/sem.h>

union semun {
  int              val;    /* Valore per SETVAL */
  struct semid_ds *buf;    /* Buffer per IPC_STAT, IPC_SET */
  unsigned short  *array;  /* Array per GETALL, SETALL */
  struct seminfo  *__buf;  /* Buffer per IPC_INFO (specifico di Linux) */
};

int ask(int* semidp, struct sembuf **sopsp);

static struct semid_ds  semid_ds;  /* stato del set di semafori */
static char error_mesg1[] = "semop: Non posso allocare spazio per un vettore
                             di %d valori.\n";
static char error_mesg2[] = "semop: Non posso allocare spazio per %d strutture 
                             sembuf. \n";

int main(int argc, char* argv[]) {
  int    i;            
  int    nsops;        /* numero di operazioni da fare */
  int    semid;        /* semid per il set di semafori */
  struct sembuf *sops; /* puntatore alle operazioni da eseguire */

  /* Cicla finche' l'utente vuole eseguire operazioni chiamando la funzione ask */
  while (nsops = ask(&semid, &sops)) {
    /* Inizializza il vettore di operazioni da eseguire */
    for (i = 0; i < nsops; i++) {
      fprintf(stderr,"\nInserire il valore per l'operazione %d di %d.\n",i+1,nsops);
      fprintf(stderr,"sem_num(i valori validi sono 0 <= sem_num < %d): ", 
              semid_ds.sem_nsems);
      scanf("%d", &sops[i].sem_num);
      fprintf(stderr, "sem_op: ");
      scanf("%d", &sops[i].sem_op);
      fprintf(stderr,"Possibili flag per sem_flg:\n");
      fprintf(stderr, "\tIPC_NOWAIT =\t%#6.6o\n",IPC_NOWAIT);
      fprintf(stderr, "\tSEM_UNDO =\t%#6.6o\n",SEM_UNDO);
      fprintf(stderr, "\tNESSUNO =\t%6d\n",0);
      fprintf(stderr, "sem_flg: ");
      /* controllare cosa fa %i su man scanf */
      scanf("%i", &sops[i].sem_flg);
    }

    /* Ricapitola la chiamata da fare a semop() */
    fprintf(stderr,"\nsemop: Chiamo semop(%d, &sops, %d) with:",semid, nsops);
    for (i = 0; i < nsops; i++) {
      fprintf(stderr, "\nsops[%d].sem_num = %d, ",i,sops[i].sem_num);
      fprintf(stderr, "sem_op = %d, ", sops[i].sem_op);
      fprintf(stderr, "sem_flg = %o\n",sops[i].sem_flg);
    }

    /* Chiama la semop() e riporta il risultato */
    if ((i = semop(semid, sops, nsops)) == -1) {
      perror("semop: semop failed");
    } else {
      fprintf(stderr, "semop: valore di ritorno = %d\n", i);
    }
  }
}


int ask(int *semidp, struct sembuf **sopsp) {
  static union semun arg;     /* argomento per semctl() */
  static int nsops = 0;       /* dimensione del vettore di sembuf */
  static int semid = -1;      /* semid del set di semafori */
  static struct sembuf *sops; /* puntatore al vettore di sembuf */
  int i;                      
 
  if (semid < 0) {
    /* Prima chiamata alla funzione ask()
       Recuperiamo semid dall'utente e lo stato corrente del set di semafori */
    fprintf(stderr,"Inserire semid del set di semafori su cui operare: ");
    scanf("%d", &semid);
    *semidp = semid;
  
    arg.buf = &semid_ds;
  
    /* chiamata a semctl() */
    if (semctl(semid, 0, IPC_STAT, arg) == -1) {
      perror("semop: semctl(IPC_STAT) fallita.");
      /* Notare che se semctl() fallisce, semid_ds viene riempita con 0, 
         e successivi test per controllare il numero di semafori
         ritorneranno 0.
         Se invece semctl() va a buon fine, arg.buf verra'
         riempito con le informazioni relative al set di semafori */

      /* allocazione della memoria per un vettore di interi la cui
         dimensione dipende dal numero di semafori inclusi nel set */  
    } else if ((arg.array = 
               (ushort *)malloc(sizeof(ushort) * semid_ds.sem_nsems)) == NULL) {
      fprintf(stderr, error_mesg1, semid_ds.sem_nsems);
      exit(1);
    }
  }
 
  /* Stampa i valori correnti dei semafori */
  if (semid_ds.sem_nsems != 0) {
    fprintf(stderr, "Ci sono %d semaphores.\n", semid_ds.sem_nsems);
    /* Chiama la funzione semctl per recuperare i valori
       di tutti i semafori del set. Nel caso di GETALL il secondo
       argomento della semctl() viene ignorato e si utilizza il
       campo array della union semun arg */
    if (semctl(semid, 0, GETALL, arg) == -1) {
      perror("semop: semctl(GETALL) fallita");
    } else {
      fprintf(stderr, "I valori correnti dei semafori sono:");
      for (i = 0; i < semid_ds.sem_nsems; i++)
        fprintf(stderr, " %d", arg.array[i]);
      fprintf(stderr, "\n");
    }
  }
  
  /* Allocazione dello spazio per le operazioni che l'utente desidera eseguire */
  fprintf(stderr, "Quante operazioni vuoi eseguire con la prossima semop()?\n");
  fprintf(stderr, "Inserire 0 or control-D per uscire: ");
  i = 0;
  if (scanf("%d", &i) == EOF || i == 0)
    exit(0);
  if (i > nsops) {
    if (nsops != 0) /* libero la memoria precedentemente allocata */
      free((char *)sops);
    nsops = i;
    /* Allocazione della memoria per le operazioni da eseguire*/
    if ((sops = 
        (struct sembuf *)malloc((nsops * sizeof(struct sembuf)))) == NULL) {
      fprintf(stderr, error_mesg2, nsops);
      exit(2);
    }
  }
  *sopsp = sops;
  return (i);
}
